#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.


class StartupInfo:
    FORCED_PROFILE_URL = "forcedProfileUrl"
    SUBSCRIBED_PRODUCTS = "subscribedProducts"
    URL = "url"

    def __init__(self, forced_profile, subscribed_products):
        self.forced_profile = forced_profile
        self.subscribed_products = subscribed_products

    def get_forced_profile_url(self) -> list:
        return self.forced_profile[self.URL]

    def get_subscribed_products_urls(self) -> list:
        return [
            product[self.URL]
            for product in self.subscribed_products
        ]

    @staticmethod
    def from_dict(data):
        return StartupInfo(
            data[StartupInfo.FORCED_PROFILE_URL],
            data[StartupInfo.SUBSCRIBED_PRODUCTS]
        )

    def __str__(self):
        return f"forced_profile: {self.forced_profile}, subscribed_products: {self.subscribed_products}"



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
from octobot_tentacles_manager.api.configurator import get_tentacles_setup_config

TEST_CONFIG_FOLDER = "tests/static"


def load_test_tentacles_config():
    return get_tentacles_setup_config(config_path=f"{TEST_CONFIG_FOLDER}/default_tentacles_config.json")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest

from exchanges_tests import abstract_authenticated_exchange_tester

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestBinanceAuthenticatedExchange(
    abstract_authenticated_exchange_tester.AbstractAuthenticatedExchangeTester
):
    # enter exchange name as a class variable here
    EXCHANGE_NAME = "binance"
    ORDER_CURRENCY = "BTC"
    SETTLEMENT_CURRENCY = "BUSD"
    SYMBOL = f"{ORDER_CURRENCY}/{SETTLEMENT_CURRENCY}"
    ORDER_SIZE = 50  # % of portfolio to include in test orders

    async def test_get_portfolio(self):
        await super().test_get_portfolio()

    async def test_create_and_cancel_limit_orders(self):
        await super().test_create_and_cancel_limit_orders()

    async def test_create_and_fill_market_orders(self):
        await super().test_create_and_fill_market_orders()

    async def test_get_my_recent_trades(self):
        await super().test_get_my_recent_trades()

    async def test_get_closed_orders(self):
        await super().test_get_closed_orders()

    async def test_create_and_cancel_stop_orders(self):
        # pass if not implemented
        pass

    async def test_edit_limit_order(self):
        # pass if not implemented
        pass

    async def test_edit_stop_order(self):
        # pass if not implemented
        pass

    async def test_create_bundled_orders(self):
        # pass if not implemented
        pass



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.community.community_supports as community_supports
import octobot.community.errors as errors


class CommunityUserAccount:
    USER_DATA_CONTENT = "content"
    BOT_DEVICE = "device"
    BOT_URLS = "urls"
    BOT_DEPLOYMENT = "deployment"
    BOT_DEPLOYMENT_TYPE = "type"
    SELF_HOSTED_BOT_DEPLOYMENT_TYPE = "self-hosted"
    METADATA = "metadata"
    FILLED_FORMS = "filledForms"
    NO_SELECTED_BOT_DESC = "No selected bot. Please select a bot to enable your community features."

    def __init__(self):
        self.gql_user_id = None
        self.gql_bot_id = None
        self.gql_access_token = None
        self.supports = community_supports.CommunitySupports()

        self._profile_raw_data = None
        self._selected_bot_raw_data = None
        self._all_user_bots_raw_data = []

    def has_user_data(self):
        return self._profile_raw_data is not None

    def has_selected_bot_data(self):
        return self._selected_bot_raw_data is not None

    def get_email(self):
        return self._profile_raw_data["email"]

    def get_user_id(self):
        return self._profile_raw_data["id"]

    def get_graph_token(self):
        return self._get_user_data_content()["graph_token"]

    def get_has_donated(self):
        return self._get_user_data_content()["has_donated"]

    def get_filled_forms_ids(self):
        return self._get_user_data_metadata().get(self.FILLED_FORMS, [])

    def get_all_user_bots_raw_data(self):
        return self._all_user_bots_raw_data

    def get_selected_bot_raw_data(self, raise_on_missing=False):
        if raise_on_missing:
            self._ensure_selected_bot_data()
        return self._selected_bot_raw_data

    def is_self_hosted(self, bot):
        deployment = bot.get(self.BOT_DEPLOYMENT)
        if not deployment:
            return True
        try:
            return deployment[self.BOT_DEPLOYMENT_TYPE] == self.SELF_HOSTED_BOT_DEPLOYMENT_TYPE
        except KeyError:
            return True

    def get_bot_deployment_url(self):
        self._ensure_selected_bot_data()
        urls = self._selected_bot_raw_data[self.BOT_DEPLOYMENT][self.BOT_URLS]
        if urls:
            return urls[0]["url"]
        raise errors.BotError("No deployment url in selected bot")

    def get_selected_bot_device_uuid(self):
        try:
            return self.get_selected_bot_raw_data(raise_on_missing=True).get(self.BOT_DEVICE, {}).get("uuid", None)
        except AttributeError:
            raise errors.NoBotDeviceError("No device associated to the select bot")

    def get_selected_bot_device_name(self):
        try:
            return self.get_selected_bot_raw_data(raise_on_missing=True).get(self.BOT_DEVICE, {}).get("name", None)
        except AttributeError:
            raise errors.NoBotDeviceError("No device associated to the select bot")

    @staticmethod
    def get_bot_id(bot):
        return bot["_id"]

    @staticmethod
    def get_bot_name_or_id(bot):
        return bot["name"] or CommunityUserAccount.get_bot_id(bot)

    def set_profile_raw_data(self, profile_raw_data):
        self._profile_raw_data = profile_raw_data

    def set_selected_bot_raw_data(self, selected_bot_raw_data):
        self._selected_bot_raw_data = selected_bot_raw_data

    def set_selected_bot_device_raw_data(self, bot_raw_data_with_device):
        device_data = bot_raw_data_with_device[self.BOT_DEVICE]
        if device_data is None:
            raise errors.NoBotDeviceError("No device associated to the select bot")
        self._selected_bot_raw_data[self.BOT_DEVICE] = device_data

    def set_all_user_bots_raw_data(self, all_user_bots_raw_data):
        self._all_user_bots_raw_data = all_user_bots_raw_data

    def _get_user_data_content(self):
        return self._profile_raw_data[self.USER_DATA_CONTENT]

    def _get_user_data_metadata(self):
        return self._profile_raw_data.get(self.METADATA, {})

    def _ensure_selected_bot_data(self):
        if self._selected_bot_raw_data is None:
            raise errors.BotError(self.NO_SELECTED_BOT_DESC)

    def flush_bot_details(self):
        self.gql_bot_id = None
        self._selected_bot_raw_data = None
        self._all_user_bots_raw_data = []

    def flush(self):
        self.gql_user_id = None
        self.gql_access_token = None
        self.supports = community_supports.CommunitySupports()
        self._profile_raw_data = None
        self.flush_bot_details()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import contextlib
import decimal

import octobot_trading.enums as trading_enums
import octobot_trading.constants as trading_constants
from exchanges_tests import abstract_authenticated_exchange_tester


class AbstractAuthenticatedFutureExchangeTester(
    abstract_authenticated_exchange_tester.AbstractAuthenticatedExchangeTester
):
    # enter exchange name as a class variable here*
    EXCHANGE_TYPE = trading_enums.ExchangeTypes.FUTURE.value
    PORTFOLIO_TYPE_FOR_SIZE = trading_constants.CONFIG_PORTFOLIO_TOTAL
    REQUIRES_SYMBOLS_TO_GET_POSITIONS = False
    INVERSE_SYMBOL = None
    MIN_PORTFOLIO_SIZE = 2  # ensure fetching currency for linear and inverse

    async def test_get_empty_linear_and_inverse_positions(self):
        # ensure fetch empty positions
        async with self.local_exchange_manager():
            await self.inner_test_get_empty_linear_and_inverse_positions()

    async def inner_test_get_empty_linear_and_inverse_positions(self):
        positions = await self.get_positions()
        self._check_position_content(positions)
        for contract_type in (trading_enums.FutureContractType.LINEAR_PERPETUAL,
                              trading_enums.FutureContractType.INVERSE_PERPETUAL):
            if not self.has_empty_position(self.get_filtered_positions(positions, contract_type)):
                empty_position_symbol = self.get_other_position_symbol(positions, contract_type)
                empty_position = await self.get_position(empty_position_symbol)
                assert self.is_position_empty(empty_position)

    def _check_position_content(self, positions):
        for position in positions:
            assert position[trading_enums.ExchangeConstantsPositionColumns.SYMBOL.value]
            # should not be 0 in octobot
            assert position[trading_enums.ExchangeConstantsPositionColumns.LEVERAGE.value] > 0

    async def inner_test_create_and_fill_market_orders(self):
        portfolio = await self.get_portfolio()
        position = await self.get_position()
        pre_order_positions = await self.get_positions()
        current_price = await self.get_price()
        price = self.get_order_price(current_price, False)
        size = self.get_order_size(portfolio, price)
        # buy: increase position
        buy_market = await self.create_market_order(current_price, size, trading_enums.TradeOrderSide.BUY)
        self.check_created_market_order(buy_market, size, trading_enums.TradeOrderSide.BUY)
        await self.wait_for_fill(buy_market)
        post_buy_portfolio = await self.get_portfolio()
        post_buy_position = await self.get_position()
        self.check_portfolio_changed(portfolio, post_buy_portfolio, False)
        self.check_position_changed(position, post_buy_position, True)
        post_order_positions = await self.get_positions()
        self.check_position_in_positions(pre_order_positions + post_order_positions)
        # sell: reset portfolio & position
        sell_market = await self.create_market_order(current_price, size, trading_enums.TradeOrderSide.SELL)
        self.check_created_market_order(sell_market, size, trading_enums.TradeOrderSide.SELL)
        await self.wait_for_fill(sell_market)
        post_sell_portfolio = await self.get_portfolio()
        post_sell_position = await self.get_position()
        self.check_portfolio_changed(post_buy_portfolio, post_sell_portfolio, True)
        self.check_position_changed(post_buy_position, post_sell_position, False)
        # position is back to what it was at the beginning on the test
        self.check_position_size(position, post_sell_position)

    async def test_create_bundled_orders(self):
        async with self.local_exchange_manager(), self.required_empty_position():
            await self.inner_test_create_bundled_orders()

    async def get_position(self, symbol=None):
        return await self.exchange_manager.exchange.get_position(symbol or self.SYMBOL)

    async def get_positions(self):
        symbols = None
        if self.REQUIRES_SYMBOLS_TO_GET_POSITIONS:
            if self.INVERSE_SYMBOL is None:
                raise AssertionError(f"INVERSE_SYMBOL is required")
            symbols = [self.SYMBOL, self.INVERSE_SYMBOL]
        return await self.exchange_manager.exchange.get_positions(symbols=symbols)

    @contextlib.asynccontextmanager
    async def required_empty_position(self):
        position = await self.get_position()
        if not self.is_position_empty(position):
            raise AssertionError(f"Empty {self.SYMBOL} position required for bundle orders tests")
        try:
            yield
        finally:
            position = await self.get_position()
            assert self.is_position_empty(position)

    async def load_contract(self, symbol=None):
        symbol = symbol or self.SYMBOL
        if self.exchange_manager.is_future and symbol not in self.exchange_manager.exchange.pair_contracts:
            await self.exchange_manager.exchange.load_pair_future_contract(symbol)

    async def enable_partial_take_profits_and_stop_loss(self, mode, symbol=None):
        await self.exchange_manager.exchange.set_symbol_partial_take_profit_stop_loss(
            symbol or self.SYMBOL, False, trading_enums.TakeProfitStopLossMode.PARTIAL)

    async def create_market_stop_loss_order(self, current_price, stop_price, size, side, symbol=None,
                                            push_on_exchange=True):
        await self.enable_partial_take_profits_and_stop_loss(trading_enums.TakeProfitStopLossMode.PARTIAL,
                                                             symbol=symbol)
        return await super().create_market_stop_loss_order(
            current_price, stop_price, size, side, symbol=symbol,
            push_on_exchange=push_on_exchange
        )

    async def create_order(self, price, current_price, size, side, order_type,
                           symbol=None, push_on_exchange=True):
        if not size:
            raise AssertionError(f"Size in required to create an order, provided size is {size}")
        # contracts are required to create orders
        await self.load_contract(symbol)
        return await super().create_order(
            price, current_price, size, side, order_type,
            symbol=symbol, push_on_exchange=push_on_exchange
        )

    def check_position_changed(self, previous_position, updated_position, has_increased, symbol=None):
        # use unified enums as it the ccxt position should have been parsed already
        previous_size = previous_position[trading_enums.ExchangeConstantsPositionColumns.SIZE.value]
        updated_size = updated_position[trading_enums.ExchangeConstantsPositionColumns.SIZE.value]
        if has_increased:
            assert updated_size > previous_size
        else:
            assert updated_size < previous_size

    def check_position_size(self, previous_position, updated_position):
        assert previous_position[trading_enums.ExchangeConstantsPositionColumns.SIZE.value] == \
               updated_position[trading_enums.ExchangeConstantsPositionColumns.SIZE.value]

    def has_empty_position(self, positions):
        for position in positions:
            if self.is_position_empty(position):
                # empty positions included in get_positions
                return True
        return False

    def get_filtered_positions(self, positions, contract_type):
        return [
            position
            for position in positions
            if position[trading_enums.ExchangeConstantsPositionColumns.CONTRACT_TYPE.value] is contract_type
        ]

    def get_other_position_symbol(self, positions_blacklist, contract_type):
        ignored_symbols = set(
            position[trading_enums.ExchangeConstantsPositionColumns.SYMBOL.value]
            for position in positions_blacklist
        )
        for symbol in self.exchange_manager.exchange.connector.client.markets:
            if symbol in ignored_symbols:
                continue
            if contract_type is trading_enums.FutureContractType.INVERSE_PERPETUAL \
               and self.exchange_manager.exchange.is_inverse_symbol(symbol):
                return symbol
            elif contract_type is trading_enums.FutureContractType.LINEAR_PERPETUAL \
                 and self.exchange_manager.exchange.is_linear_symbol(symbol):
                return symbol
        raise AssertionError(f"No free symbol for {contract_type}")

    def is_position_empty(self, position):
        return position[trading_enums.ExchangeConstantsPositionColumns.SIZE.value] == trading_constants.ZERO

    def check_position_in_positions(self, positions, symbol=None):
        symbol = symbol or self.SYMBOL
        for position in positions:
            if position[trading_enums.ExchangeConstantsPositionColumns.SYMBOL.value] == symbol:
                return True
        raise AssertionError(f"Can't find position for symbol: {symbol}")

    async def order_in_open_orders(self, previous_open_orders, order):
        open_orders = await self.get_open_orders()
        assert len(open_orders) == len(previous_open_orders) + 1
        for open_order in open_orders:
            if open_order[trading_enums.ExchangeConstantsOrderColumns.ID.value] == order.order_id:
                return True
        return False

    def check_theoretical_cost(self, symbol, quantity, price, cost):
        if symbol.is_inverse():
            theoretical_cost = quantity
        else:
            theoretical_cost = quantity * price
        assert theoretical_cost * decimal.Decimal("0.8") <= cost <= theoretical_cost * decimal.Decimal("1.2")

    def _get_all_symbols(self):
        return [self.SYMBOL, self.INVERSE_SYMBOL]



# cython: language_level=3
#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot.strategy_optimizer cimport test_suite_result
from octobot.strategy_optimizer.test_suite_result cimport (
    TestSuiteResult,
    TestSuiteResultSummary,
)
from octobot.strategy_optimizer cimport strategy_optimizer
from octobot.strategy_optimizer.strategy_optimizer cimport (
    StrategyOptimizer,
)
from octobot.strategy_optimizer cimport strategy_test_suite
from octobot.strategy_optimizer.strategy_test_suite cimport (
    StrategyTestSuite,
)

__all__ = [
    "TestSuiteResult",
    "TestSuiteResultSummary",
    "StrategyOptimizer",
    "StrategyTestSuite",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import base64
import contextlib
import json
import time
import datetime

import requests
import aiohttp

import octobot.constants as constants
import octobot.community.errors as errors
import octobot.community.identifiers_provider as identifiers_provider
import octobot.community.community_supports as community_supports
import octobot.community.startup_info as startup_info
import octobot.community.feeds as community_feeds
import octobot.community.graphql_requests as graphql_requests
import octobot.community.community_user_account as community_user_account
import octobot_commons.constants as commons_constants
import octobot_commons.authentication as authentication
import octobot_commons.configuration as commons_configuration


def _selected_bot_update(func):
    async def wrapper(*args, **kwargs):
        self = args[0]
        await self.gql_login_if_required()
        if self._update_bot_lock is None:
            self._update_bot_lock = asyncio.Lock()
        async with self._update_bot_lock:
            self.logger.debug(f"@selected_bot_update: entering {func.__name__}")
            updated_bot = await func(*args, **kwargs)
        self.logger.debug(f"@selected_bot_update: exited {func.__name__}")
        self.user_account.set_selected_bot_raw_data(updated_bot)
        return updated_bot

    return wrapper


class CommunityAuthentication(authentication.Authenticator):
    """
    Authentication utility
    """
    ALLOWED_TIME_DELAY = 1 * commons_constants.MINUTE_TO_SECONDS
    NEW_ACCOUNT_INITIALIZE_TIMEOUT = 1 * commons_constants.MINUTE_TO_SECONDS
    LOGIN_TIMEOUT = 20
    BOT_NOT_FOUND_RETRY_DELAY = 1
    AUTHORIZATION_HEADER = "authorization"
    SESSION_HEADER = "X-Session"
    GQL_AUTHORIZATION_HEADER = "Authorization"

    def __init__(self, feed_url, config=None):
        super().__init__()
        self.authentication_url = identifiers_provider.IdentifiersProvider.BACKEND_AUTH_URL
        self.feed_url = feed_url
        self.edited_config = config
        self.initialized_event = None

        self.user_account = community_user_account.CommunityUserAccount()
        self._community_token = self._get_encoded_community_token()
        self._auth_token = None
        self._backend_session = requests.Session()
        self._aiohttp_gql_session = None
        self._aiohttp_backend_session = None
        self._fetch_account_task = None
        self._restart_task = None
        self._community_feed = None
        self._login_completed = None
        self._update_bot_lock = None
        self._startup_info = None

        self._update_sessions_headers()

    @staticmethod
    def create(configuration: commons_configuration.Configuration):
        return CommunityAuthentication.instance(
            identifiers_provider.IdentifiersProvider.FEED_URL,
            config=configuration,
        )

    def update(self, configuration: commons_configuration.Configuration):
        self.edited_config = configuration
        self._update_sessions_headers()

    def get_logged_in_email(self):
        if self.user_account.has_user_data():
            return self.user_account.get_email()
        raise authentication.AuthenticationRequired()

    def get_packages(self):
        try:
            #TODO
            return []
        except json.JSONDecodeError:
            return []

    def is_feed_connected(self):
        return self._community_feed is not None and self._community_feed.is_connected_to_remote_feed()

    def get_feed_last_message_time(self):
        if self._community_feed is None:
            return None
        return self._community_feed.last_message_time

    def has_filled_form(self, form_id):
        if not self.user_account.has_user_data():
            raise authentication.AuthenticationRequired()
        return form_id in self.user_account.get_filled_forms_ids()

    async def register_filled_form(self, form_id):
        if self.has_filled_form(form_id):
            return
        updated_filled_forms = self.user_account.get_filled_forms_ids()
        updated_filled_forms.append(form_id)
        await self._update_account_metadata({
            self.user_account.FILLED_FORMS: updated_filled_forms
        })

    def get_user_id(self):
        if not self.user_account.has_user_data():
            raise authentication.AuthenticationRequired()
        return self.user_account.get_user_id()

    def get_deployment_url(self):
        return self.user_account.get_bot_deployment_url()

    def get_is_signal_receiver(self):
        if self._community_feed is None:
            return False
        return self._community_feed.is_signal_receiver

    def get_is_signal_emitter(self):
        if self._community_feed is None:
            return False
        return self._community_feed.is_signal_emitter

    def get_signal_community_url(self, signal_identifier):
        return f"{identifiers_provider.IdentifiersProvider.COMMUNITY_URL}/product/{signal_identifier}"

    async def update_supports(self):
        self._update_supports(200, self._supports_mock())
        return
        # TODO use real support fetch when implemented
        async with self._aiohttp_gql_session.get("supports_url") as resp:
            self._update_supports(resp.status, await resp.json())

    def ensure_async_loop(self):
        # elements should be bound to the current loop
        if self._aiohttp_gql_session is not None and self._aiohttp_gql_session.loop is not asyncio.get_event_loop():
            self._aiohttp_gql_session = None
        if self._aiohttp_backend_session is not None and self._aiohttp_backend_session.loop is not asyncio.get_event_loop():
            self._aiohttp_backend_session = None
        if self.initialized_event is not None and self.initialized_event._loop is not asyncio.get_event_loop():
            should_set = self.initialized_event.is_set()
            self.initialized_event = asyncio.Event()
            if should_set:
                self.initialized_event.set()

    def is_initialized(self):
        return self.initialized_event is not None and self.initialized_event.is_set()

    def init_account(self):
        self.initialized_event = asyncio.Event()
        self._fetch_account_task = asyncio.create_task(self._auth_and_fetch_account())

    async def async_init_account(self):
        self.init_account()
        await self._fetch_account_task

    async def _ensure_bot_device(self):
        try:
            self.user_account.get_selected_bot_device_uuid()
        except errors.NoBotDeviceError:
            bot_id = self.user_account.get_bot_id(self.user_account.get_selected_bot_raw_data())
            self.logger.info(f"Creating a bot device for current bot with id: {bot_id} (no associated bot device)")
            self.user_account.set_selected_bot_device_raw_data(await self._create_new_bot_device(bot_id))

    async def _create_community_feed_if_necessary(self) -> bool:
        if self._community_feed is None:
            self._community_feed = community_feeds.community_feed_factory(
                self.feed_url,
                self,
                constants.COMMUNITY_FEED_DEFAULT_TYPE
            )
            return True
        return False

    async def _ensure_init_community_feed(self):
        await self._create_community_feed_if_necessary()
        if not self._community_feed.is_connected() and self._community_feed.can_connect():
            if self.initialized_event is not None and not self.initialized_event.is_set():
                await asyncio.wait_for(self.initialized_event.wait(), self.LOGIN_TIMEOUT)
            if not self.is_logged_in():
                raise authentication.AuthenticationRequired("You need to be authenticated to be able to "
                                                            "connect to signals")
            await self._ensure_bot_device()
            await self._community_feed.start()

    async def register_feed_callback(self, channel_type, callback, identifier=None):
        try:
            await self._ensure_init_community_feed()
            await self._community_feed.register_feed_callback(channel_type, callback, identifier=identifier)
        except errors.BotError as e:
            self.logger.error(f"Impossible to connect to community signals: {e}")

    async def send(self, message, channel_type, identifier=None):
        """
        Sends a message
        """
        await self._ensure_init_community_feed()
        await self._community_feed.send(message, channel_type, identifier)

    @staticmethod
    def _build_gql_request_body(query, variables, operation_name):
        request_body = {
            "query": query
        }
        if variables is not None:
            request_body["variables"] = variables
        if operation_name is not None:
            request_body["operationName"] = operation_name
        return request_body

    async def async_graphql_query(self, query, query_name, variables=None, operation_name=None,
                                  expected_code=None, allow_retry_on_expired_token=True):
        try:
            async with self._authenticated_qgl_session() as session:
                t0 = time.time()
                self.logger.debug(f"starting {query_name} graphql query")
                resp = await session.post(
                    f"{identifiers_provider.IdentifiersProvider.GQL_BACKEND_API_URL}",
                    json=self._build_gql_request_body(query, variables, operation_name)
                )
                self.logger.debug(f"graphql query {query_name} done in {time.time() - t0} seconds")
                if resp.status == 401:
                    # access token expired
                    raise authentication.AuthenticationRequired
                json_resp = await resp.json()
                if errs := json_resp.get("errors"):
                    raise errors.RequestError(f"Error when running graphql query [{query_name}]: "
                                              f"{errs[0].get('message', errs)}")
                if expected_code is None or resp.status == expected_code:
                    return json_resp["data"][query_name]
                raise errors.StatusCodeRequestError(
                    f"Wrong status code running graphql query [{query_name}]: expected {expected_code}, "
                    f"got: {resp.status}. Text: {await resp.text()}")
        except authentication.AuthenticationRequired:
            if allow_retry_on_expired_token:
                return await self.async_graphql_query(
                    query, query_name, variables=variables, operation_name=operation_name,
                    expected_code=expected_code, allow_retry_on_expired_token=False)
            else:
                raise

    @contextlib.asynccontextmanager
    async def _authenticated_qgl_session(self):
        await self.gql_login_if_required()
        session = self.get_gql_aiohttp_session()
        try:
            yield session
        except authentication.AuthenticationRequired:
            # reset token to force re-login
            self.user_account.gql_access_token = None
            if CommunityAuthentication.GQL_AUTHORIZATION_HEADER in session.headers:
                session.headers.pop(CommunityAuthentication.GQL_AUTHORIZATION_HEADER)
            raise

    async def wait_for_login_if_processing(self):
        if self._login_completed is not None:
            # ensure login details have been fetched
            await asyncio.wait_for(self._login_completed.wait(), self.LOGIN_TIMEOUT)

    async def gql_login_if_required(self):
        await self.wait_for_login_if_processing()
        if self.user_account.gql_access_token is not None:
            return
        try:
            token = self.user_account.get_graph_token()
        except (KeyError, TypeError) as e:
            raise authentication.AuthenticationRequired("Authentication required") from e
        async with self.get_aiohttp_session().post(
                identifiers_provider.IdentifiersProvider.GQL_AUTH_URL, json={"key": token}
        ) as resp:
            json_resp = await resp.json()
            if resp.status == 200:
                self.user_account.gql_access_token = json_resp["access_token"]
                self.user_account.gql_user_id = json_resp["user_id"]
                self._update_sessions_headers()
            else:
                raise authentication.FailedAuthentication(f"Failed to authenticate to graphql server: "
                                                          f"status: {resp.status}, data: {json_resp}")

    def can_authenticate(self):
        return "todo" not in self.authentication_url    # pylint: disable=E1135

    def must_be_authenticated_through_authenticator(self):
        return constants.IS_CLOUD_ENV

    async def login(self, email, password, password_token=None):
        self._ensure_email(email)
        self._ensure_community_url()
        self._reset_tokens()
        params = {
            "email": email,
        }
        if password_token:
            params["password_token"] = password_token
        else:
            params["password"] = password
        resp = self._backend_session.post(self.authentication_url, json=params)
        try:
            self._handle_auth_result(resp.status_code, resp.json(), resp.headers)
        except json.JSONDecodeError as e:
            raise authentication.FailedAuthentication(e)
        if self.is_logged_in():
            if self.initialized_event is None:
                self.initialized_event = asyncio.Event()
            await self._on_authenticated()
            self.initialized_event.set()

    async def register(self, email, password):
        if self.must_be_authenticated_through_authenticator():
            raise authentication.AuthenticationError("Creating a new account is not authorized on this environment.")
        # always logout before creating a new account
        self.logout()
        self._ensure_community_url()
        params = {
            "email": email,
            "password": password,
            community_user_account.CommunityUserAccount.METADATA: {
                "hasRegisteredFromSelfHosted": True,
            }
        }
        async with self.get_aiohttp_session().post(
                identifiers_provider.IdentifiersProvider.BACKEND_ACCOUNT_URL, json=params
        ) as resp:
            try:
                self._handle_register_result(resp.status, await resp.json(), resp.headers)
            except json.JSONDecodeError as e:
                raise authentication.FailedAuthentication(e)
        if self.is_logged_in():
            await self._on_register()

    async def _update_account_metadata(self, metadata_update):
        params = {
            community_user_account.CommunityUserAccount.METADATA: metadata_update
        }
        async with self.get_aiohttp_session().put(
                identifiers_provider.IdentifiersProvider.BACKEND_ACCOUNT_URL, json=params
        ) as resp:
            self._handle_account_update_result(resp.status, await resp.json(), resp.headers)

    async def _on_register(self):
        if self.initialized_event is None:
            self.initialized_event = asyncio.Event()
        t0 = time.time()
        try:
            while not self._is_new_account_fully_initialized() \
                    and time.time() - t0 < self.NEW_ACCOUNT_INITIALIZE_TIMEOUT:
                await self._async_check_auth()
        finally:
            if self._is_new_account_fully_initialized():
                self.logger.debug(f"New account has been fully initialized, now processing authenticated checkups")
                await self._on_authenticated()
                self.logger.debug(f"Authenticated checkups complete")
            else:
                self.logger.error(f"New account has not been fully initialized")
            self.initialized_event.set()

    def _is_new_account_fully_initialized(self):
        # required elements to be fully initialized
        try:
            self.user_account.get_graph_token()
        except (KeyError, TypeError):
            return False
        return True

    async def update_selected_bot(self):
        self.user_account.flush_bot_details()
        await self._load_bot_if_selected()
        if not self.user_account.has_selected_bot_data():
            self.logger.info(self.user_account.NO_SELECTED_BOT_DESC)

    async def _load_bot_if_selected(self):
        # 1. use user selected bot id if any
        if saved_uuid := self._get_saved_gql_bot_id():
            try:
                await self.select_bot(saved_uuid)
                return
            except errors.BotNotFoundError as e:
                # proceed to 2.
                self.logger.warning(str(e))
        # 2. fetch all user bots and create one if none, otherwise ask use for which one to use
        await self.load_user_bots()
        if len(self.user_account.get_all_user_bots_raw_data()) == 0:
            await self.select_bot(
                self.user_account.get_bot_id(
                    await self.create_new_bot()
                )
            )
        # more than one possible bot, can't auto-select one

    async def select_bot(self, bot_id):
        fetched_bot = await self._fetch_bot(bot_id)
        if fetched_bot is None:
            # retry after some time, if still None, there is an issue
            await asyncio.sleep(self.BOT_NOT_FOUND_RETRY_DELAY)
            fetched_bot = await self._fetch_bot(bot_id)
        if fetched_bot is None:
            raise errors.BotNotFoundError(f"Can't find bot with id: {bot_id}")
        self.user_account.set_selected_bot_raw_data(fetched_bot)
        bot_name = self.user_account.get_bot_name_or_id(self.user_account.get_selected_bot_raw_data())
        self.logger.debug(f"Selected bot '{bot_name}'")
        self.user_account.gql_bot_id = bot_id
        self._save_gql_bot_id(self.user_account.gql_bot_id)
        await self.on_new_bot_select()

    async def load_user_bots(self):
        self.user_account.set_all_user_bots_raw_data(
            self._get_self_hosted_bots(
                await self._fetch_bots()
            )
        )

    async def get_startup_info(self):
        if self._startup_info is None:
            if self.user_account.gql_bot_id is None:
                raise errors.BotError("No selected bot")
            self._startup_info = startup_info.StartupInfo.from_dict(
                await self._fetch_startup_info(
                    self.user_account.gql_bot_id
                )
            )
        return self._startup_info

    async def get_subscribed_profile_urls(self):
        subscribed_profiles = await self._fetch_subscribed_profiles()
        return [
            profile_data["url"]
            for profile_data in subscribed_profiles["data"]
        ]

    async def update_trades(self, trades: list):
        """
        Updates authenticated account trades
        """
        if not self.is_logged_in():
            return
        try:
            formatted_trades = [
                {
                    "date": self._get_graphql_formatted_time(trade.executed_time),
                    "exchange": trade.exchange_manager.exchange_name,
                    "price": str(trade.executed_price),
                    "quantity": str(trade.executed_quantity),
                    "symbol": trade.symbol,
                    "type": trade.trade_type.value,
                }
                for trade in trades
            ]
            await self._update_bot_trades(formatted_trades)
        except Exception as err:
            self.logger.exception(err, True, f"Error when updating community trades {err}")

    async def update_portfolio(self, current_value: dict, initial_value: dict,
                               unit: str, content: dict, history: dict, price_by_asset: dict):
        """
        Updates authenticated account portfolio
        """
        if not self.is_logged_in():
            return
        try:
            ref_market_current_value = current_value[unit]
            ref_market_initial_value = initial_value[unit]
            formatted_content = [
                {
                    "asset": key,
                    "quantity": str(quantity[commons_constants.PORTFOLIO_TOTAL]),
                    "value": str(quantity[commons_constants.PORTFOLIO_TOTAL] * float(price_by_asset.get(key, 0))),
                }
                for key, quantity in content.items()
            ]
            formatted_history = []
            try:
                formatted_history = [
                    {
                        "date": self._get_graphql_formatted_time(timestamp),
                        "value": str(value[unit])
                    }
                    for timestamp, value in history.items()
                    if unit in value
                ]
            except KeyError:
                pass
            if not formatted_history:
                return
            await self._update_bot_portfolio(
                ref_market_current_value, ref_market_initial_value, unit,
                formatted_content, formatted_history
            )
        except Exception as err:
            self.logger.exception(err, True, f"Error when updating community portfolio {err}")

    def _get_self_hosted_bots(self, bots):
        return [
            bot
            for bot in bots
            if self.user_account.is_self_hosted(bot)
        ]

    async def on_new_bot_select(self):
        await self._update_feed_device_uuid_and_restart_feed_if_necessary()

    async def _fetch_startup_info(self, bot_id):
        return await self._execute_request(graphql_requests.select_startup_info_query, bot_id)

    async def _fetch_subscribed_profiles(self):
        return await self._execute_request(graphql_requests.select_subscribed_profiles_query)

    async def _fetch_bots(self):
        return await self._execute_request(graphql_requests.select_bots_query)

    async def _fetch_bot(self, bot_id):
        return await self._execute_request(graphql_requests.select_bot_query, bot_id)

    async def create_new_bot(self):
        await self.gql_login_if_required()
        return await self._execute_request(graphql_requests.create_bot_query, not constants.IS_CLOUD_ENV)

    async def _create_new_bot_device(self, bot_id):
        await self._execute_request(graphql_requests.create_bot_device_query, bot_id)
        # issue with createBotDevice not always returning the created device, fetch bot again to fetch device with it
        return await self._fetch_bot(bot_id)

    async def _update_feed_device_uuid_and_restart_feed_if_necessary(self):
        if self._community_feed is None or not self.initialized_event.is_set():
            # only create a new community feed if necessary
            return
        if not (self._community_feed.is_using_bot_device(self.user_account) and self._community_feed.is_connected()):
            # Need to connect using the new uuid.

            # Reset restart task if running
            if self._restart_task is not None and not self._restart_task.done():
                self._restart_task.cancel()
                self._community_feed.remove_device_details()
            await self._ensure_bot_device()
            self._restart_task = asyncio.create_task(self._community_feed.restart())

    @_selected_bot_update
    async def update_bot_config_and_stats(self, profile_name, profitability):
        return await self._execute_request(
            graphql_requests.update_bot_config_and_stats_query,
            self.user_account.gql_bot_id,
            profile_name,
            profitability
        )

    @_selected_bot_update
    async def _update_bot_trades(self, trades):
        return await self._execute_request(
            graphql_requests.update_bot_trades_query,
            self.user_account.gql_bot_id,
            trades
        )

    @_selected_bot_update
    async def _update_bot_portfolio(self, current_value, initial_value, unit, content, history):
        return await self._execute_request(
            graphql_requests.update_bot_portfolio_query,
            self.user_account.gql_bot_id,
            current_value, initial_value, unit, content, history
        )

    async def _execute_request(self, request_factory, *args, **kwargs):
        query, variables, query_name = request_factory(*args, **kwargs)
        return await self.async_graphql_query(query, query_name, variables=variables, expected_code=200)

    def logout(self):
        """
        logout and remove saved auth details
        Warning: also call stop_feeds if feeds have to be stopped (not done here to keep method sync)
        """
        self._reset_tokens()
        self.remove_login_detail()
        for task in (self._restart_task, self._fetch_account_task):
            if task is not None and not task.done():
                task.cancel()
        self._restart_task = self._fetch_account_task = None
        if self._community_feed is not None:
            self._community_feed.remove_device_details()

    async def stop_feeds(self):
        if self._community_feed is not None and self._community_feed.is_connected():
            await self._community_feed.stop()

    def is_logged_in(self):
        return bool(self._auth_token and self.user_account.has_user_data())

    def has_login_info(self):
        return bool(self._get_saved_token())

    def ensure_token_validity(self):
        if not self.is_logged_in():
            # try to login with saved token if available
            self._try_auto_login()
            if not self.is_logged_in():
                # still not logged in: raise
                raise authentication.AuthenticationRequired()

    def remove_login_detail(self):
        self.user_account.flush()
        self._save_login_token("")
        self._save_gql_bot_id("")
        self.logger.debug("Removed community login data")

    def get_aiohttp_session(self):
        if self._aiohttp_backend_session is None:
            self._aiohttp_backend_session = aiohttp.ClientSession()
            self._update_sessions_headers()
        return self._aiohttp_backend_session

    def get_gql_aiohttp_session(self):
        if self._aiohttp_gql_session is None:
            self._aiohttp_gql_session = aiohttp.ClientSession()
            self._update_sessions_headers()
        return self._aiohttp_gql_session

    async def stop(self):
        self.logger.debug("Stopping ...")
        await self.stop_feeds()
        if self._fetch_account_task is not None and not self._fetch_account_task.done():
            self._fetch_account_task.cancel()
        if self._restart_task is not None and not self._restart_task.done():
            self._restart_task.cancel()
        if self._aiohttp_backend_session is not None:
            await self._aiohttp_backend_session.close()
        if self._aiohttp_gql_session is not None:
            await self._aiohttp_gql_session.close()
        self.logger.debug("Stopped")

    def _update_supports(self, resp_status, json_data):
        if resp_status == 200:
            self.user_account.supports = community_supports.CommunitySupports.from_community_dict(json_data)
            self.logger.debug(f"Fetched supports data.")
        else:
            self.logger.error(f"Error when fetching community support, "
                              f"error code: {resp_status}")

    def _supports_mock(self):
        return {
            "data": {
                "attributes": {
                    "support_role": self._get_support_role()
                }
            }
        }

    def _get_support_role(self):
        try:
            if self.user_account.get_has_donated():
                return community_supports.CommunitySupports.OCTOBOT_DONOR_ROLE
        except KeyError:
            pass
        return community_supports.CommunitySupports.DEFAULT_SUPPORT_ROLE

    async def _on_authenticated(self):
        await self.update_supports()
        await self.update_selected_bot()

    async def _auth_and_fetch_account(self):
        try:
            await self._async_try_auto_login()
            if not self.is_logged_in():
                return
            await self._on_authenticated()
        except authentication.UnavailableError as e:
            self.logger.exception(e, True, f"Error when fetching community supports, "
                                           f"please check your internet connection.")
        except Exception as e:
            self.logger.exception(e, True, f"Error when fetching community supports: {e}({e.__class__.__name__})")
        finally:
            self.initialized_event.set()

    def _save_login_token(self, value):
        self._save_value_in_config(constants.CONFIG_COMMUNITY_TOKEN, value)

    def _save_gql_bot_id(self, gql_bot_id):
        self._save_value_in_config(constants.CONFIG_COMMUNITY_BOT_ID, gql_bot_id)

    def _get_saved_token(self):
        return self._get_value_in_config(constants.CONFIG_COMMUNITY_TOKEN)

    def _get_saved_gql_bot_id(self):
        return constants.COMMUNITY_BOT_ID or self._get_value_in_config(constants.CONFIG_COMMUNITY_BOT_ID)

    def _save_value_in_config(self, key, value):
        if self.edited_config is not None:
            if constants.CONFIG_COMMUNITY not in self.edited_config.config:
                self.edited_config.config[constants.CONFIG_COMMUNITY] = {}
            self.edited_config.config[constants.CONFIG_COMMUNITY][key] = value
            self.edited_config.save()

    def _get_value_in_config(self, key):
        if self.edited_config is not None:
            try:
                return self.edited_config.config[constants.CONFIG_COMMUNITY][key]
            except KeyError:
                return ""
        return None

    def _try_auto_login(self):
        self._auth_token = self._get_saved_token()
        if self._auth_token:
            # try to login using config data
            self._auto_login()

    async def _async_try_auto_login(self):
        self._auth_token = self._get_saved_token()
        if self._auth_token:
            # try to login using config data
            await self._async_auto_login()

    @contextlib.contextmanager
    def _auth_handler(self):
        try:
            yield
        except authentication.FailedAuthentication as e:
            self.logger.warning(f"Invalid authentication details, please re-authenticate. {e}")
            self.logout()
        except authentication.UnavailableError:
            raise
        except Exception as e:
            self.logger.exception(e, True, f"Error when trying to refresh community login: {e}")

    def _auto_login(self):
        with self._auth_handler():
            self._check_auth()

    async def _async_auto_login(self):
        with self._auth_handler():
            await self._async_check_auth()

    @contextlib.contextmanager
    def _auth_context(self):
        self._ensure_community_url()
        try:
            yield
        except (requests.ConnectionError, aiohttp.ClientConnectionError) as e:
            raise authentication.UnavailableError from e

    def _check_auth(self):
        with self._auth_context():
            with self._backend_session.get(identifiers_provider.IdentifiersProvider.BACKEND_ACCOUNT_URL) as resp:
                self._handle_auth_result(resp.status_code, resp.json(), resp.headers)

    async def _async_check_auth(self):
        if self._login_completed is None:
            self._login_completed = asyncio.Event()
        self._login_completed.clear()
        with self._auth_context():
            async with self.get_aiohttp_session().get(
                    identifiers_provider.IdentifiersProvider.BACKEND_ACCOUNT_URL
            ) as resp:
                try:
                    self._handle_auth_result(resp.status, await resp.json(), resp.headers)
                finally:
                    self._login_completed.set()

    def _ensure_email(self, email):
        if constants.USER_ACCOUNT_EMAIL and email != constants.USER_ACCOUNT_EMAIL:
            raise authentication.AuthenticationError("The given email doesn't match the expected user email.")

    def _ensure_community_url(self):
        if not self.can_authenticate():
            raise authentication.UnavailableError("Community url required")

    def _handle_auth_result(self, status_code, json_resp, reps_headers):
        if status_code == 200 and json_resp is not None:
            if new_token := reps_headers.get(self.SESSION_HEADER):
                self._auth_token = new_token
                self._save_login_token(self._auth_token)
                self._update_sessions_headers()
            self.user_account.set_profile_raw_data(json_resp)
            return
        elif json_resp is None and status_code < 500:
            if json_resp is not None:
                if error := json_resp.get("error", None):
                    raise authentication.FailedAuthentication(f"Error when authenticating: {error['message']}")
            raise authentication.FailedAuthentication("Invalid username or password." if self._auth_token is None else
                                                      "Token expired. Please re-login to your community account")
        elif json_resp and status_code == 400:
            if error := json_resp.get("error", None):
                if "logged in" in error['message'].lower():
                    raise authentication.FailedAuthentication(f"Error when authenticating, please "
                                                              f"re-login to your community account")
        raise authentication.AuthenticationError(f"Error code: {status_code}")

    def _handle_register_result(self, status_code, json_resp, reps_headers):
        if status_code == 200 and json_resp is not None:
            if "id" in json_resp:
                # successfully created account
                if new_token := reps_headers.get(self.SESSION_HEADER):
                    self._auth_token = new_token
                    self._save_login_token(self._auth_token)
                    self._update_sessions_headers()
                self.user_account.set_profile_raw_data(json_resp)
                return
            # error on create account
            error_messages = [
                f"{field}: {details['message']}"
                for field, details in json_resp.items()
            ]
            raise authentication.AuthenticationError(", ".join(error_messages))
        raise authentication.AuthenticationError(f"Unexpected error when creating account: code: {status_code} ({json_resp})")

    def _handle_account_update_result(self, status_code, json_resp, reps_headers):
        if status_code == 200 and json_resp is not None:
            if "id" in json_resp:
                # successfully updated account
                self.user_account.set_profile_raw_data(json_resp)
                return
            # error on update account
            error_messages = [
                f"{field}: {details['message']}"
                for field, details in json_resp.items()
            ]
            raise authentication.AccountUpdateError(", ".join(error_messages))
        raise authentication.AccountUpdateError(f"Unexpected error when updating account: code: {status_code} ({json_resp})")

    def _update_sessions_headers(self):
        backend_headers = self.get_backend_headers()
        self._backend_session.headers.update(backend_headers)
        if self._aiohttp_backend_session is not None:
            self._aiohttp_backend_session.headers.update(backend_headers)
        if self._aiohttp_gql_session is not None:
            self._aiohttp_gql_session.headers.update(self.get_gql_headers())

    def get_backend_headers(self):
        headers = {
            CommunityAuthentication.AUTHORIZATION_HEADER: f"Basic {self._community_token}",
        }
        if self._auth_token is not None:
            headers[CommunityAuthentication.SESSION_HEADER] = self._auth_token
        return headers

    def get_gql_headers(self):
        headers = {}
        if self.user_account.gql_access_token is not None:
            headers[CommunityAuthentication.GQL_AUTHORIZATION_HEADER] = f"Bearer {self.user_account.gql_access_token}"
        return headers

    def _reset_tokens(self):
        self._auth_token = None
        self.user_account.flush()
        self._backend_session.headers.pop(CommunityAuthentication.SESSION_HEADER, None)
        if self._aiohttp_backend_session is not None:
            self._aiohttp_backend_session.headers.pop(CommunityAuthentication.SESSION_HEADER, None)
        if self._aiohttp_gql_session is not None:
            self._aiohttp_gql_session.headers.pop(CommunityAuthentication.GQL_AUTHORIZATION_HEADER, None)

    @staticmethod
    def _get_encoded_community_token():
        return base64.encodebytes(identifiers_provider.IdentifiersProvider.BACKEND_PUBLIC_TOKEN.encode()).decode().strip()

    @staticmethod
    def _get_graphql_formatted_time(timestamp):
        return f"{datetime.datetime.utcfromtimestamp(timestamp).isoformat('T')}Z"



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import contextlib
import decimal
import time

import octobot_commons.constants as constants
import octobot_commons.symbols as symbols
import octobot_trading.enums as trading_enums
import octobot_trading.constants as trading_constants
import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders as personal_data_orders
import octobot_tentacles_manager.api as tentacles_manager_api
from exchanges_tests import get_authenticated_exchange_manager

# always import and load tentacles
import tentacles
tentacles_manager_api.reload_tentacle_info()


class AbstractAuthenticatedExchangeTester:
    # enter exchange name as a class variable here
    EXCHANGE_NAME = None
    CREDENTIALS_EXCHANGE_NAME = None
    EXCHANGE_TENTACLE_NAME = None
    EXCHANGE_TYPE = trading_enums.ExchangeTypes.SPOT.value
    ORDER_CURRENCY = "BTC"
    SETTLEMENT_CURRENCY = "USDT"
    SYMBOL = f"{ORDER_CURRENCY}/{SETTLEMENT_CURRENCY}"
    ORDER_SIZE = 10  # % of portfolio to include in test orders
    PORTFOLIO_TYPE_FOR_SIZE = trading_constants.CONFIG_PORTFOLIO_FREE
    CONVERTS_ORDER_SIZE_BEFORE_PUSHING_TO_EXCHANGES = False
    ORDER_PRICE_DIFF = 20  # % of price difference compared to current price for limit and stop orders
    NO_FEE_ON_GET_CLOSED_ORDERS = False
    OPEN_ORDERS_IN_CLOSED_ORDERS = False
    MARKET_FILL_TIMEOUT = 15
    OPEN_TIMEOUT = 15
    CANCEL_TIMEOUT = 15
    EDIT_TIMEOUT = 15
    MIN_PORTFOLIO_SIZE = 1

    # Implement all "test_[name]" methods, call super() to run the test, pass to ignore it.
    # Override the "inner_test_[name]" method to override a test content.
    # Add method call to subclasses to be able to run them independently

    async def test_get_portfolio(self):
        async with self.local_exchange_manager():
            await self.inner_test_get_portfolio()

    async def inner_test_get_portfolio(self):
        self.check_portfolio_content(await self.get_portfolio())

    async def test_create_and_cancel_limit_orders(self):
        async with self.local_exchange_manager():
            await self.inner_test_create_and_cancel_limit_orders()

    async def inner_test_create_and_cancel_limit_orders(self):
        price = self.get_order_price(await self.get_price(), False)
        size = self.get_order_size(await self.get_portfolio(), price)
        open_orders = await self.get_open_orders()
        buy_limit = await self.create_limit_order(price, size, trading_enums.TradeOrderSide.BUY)
        self.check_created_limit_order(buy_limit, price, size, trading_enums.TradeOrderSide.BUY)
        assert await self.order_in_open_orders(open_orders, buy_limit)
        await self.cancel_order(buy_limit)
        assert await self.order_not_in_open_orders(open_orders, buy_limit)

    async def test_create_and_fill_market_orders(self):
        async with self.local_exchange_manager():
            await self.inner_test_create_and_fill_market_orders()

    async def inner_test_create_and_fill_market_orders(self):
        portfolio = await self.get_portfolio()
        current_price = await self.get_price()
        price = self.get_order_price(current_price, False)
        size = self.get_order_size(portfolio, price)
        buy_market = await self.create_market_order(current_price, size, trading_enums.TradeOrderSide.BUY)
        self.check_created_market_order(buy_market, size, trading_enums.TradeOrderSide.BUY)
        await self.wait_for_fill(buy_market)
        sell_size = self.get_sell_size_from_buy_order(buy_market)
        post_buy_portfolio = await self.get_portfolio()
        self.check_portfolio_changed(portfolio, post_buy_portfolio, False)
        # sell: reset portfolio
        sell_market = await self.create_market_order(current_price, sell_size, trading_enums.TradeOrderSide.SELL)
        self.check_created_market_order(sell_market, sell_size, trading_enums.TradeOrderSide.SELL)
        await self.wait_for_fill(sell_market)
        post_sell_portfolio = await self.get_portfolio()
        self.check_portfolio_changed(post_buy_portfolio, post_sell_portfolio, True)

    async def test_create_and_cancel_stop_orders(self):
        # pass if not implemented
        async with self.local_exchange_manager():
            await self.inner_test_create_and_cancel_stop_orders()

    async def inner_test_create_and_cancel_stop_orders(self):
        current_price = await self.get_price()
        price = self.get_order_price(current_price, False)
        size = self.get_order_size(await self.get_portfolio(), price)
        open_orders = await self.get_open_orders()
        stop_loss = await self.create_market_stop_loss_order(current_price, price, size,
                                                             trading_enums.TradeOrderSide.SELL)
        self.check_created_stop_order(stop_loss, price, size, trading_enums.TradeOrderSide.SELL)
        stop_loss_from_get_order = await self.get_order(stop_loss.order_id)
        self.check_created_stop_order(stop_loss_from_get_order, price, size, trading_enums.TradeOrderSide.SELL)
        assert await self.order_in_open_orders(open_orders, stop_loss)
        await self.cancel_order(stop_loss)
        assert await self.order_not_in_open_orders(open_orders, stop_loss)

    async def test_get_my_recent_trades(self):
        async with self.local_exchange_manager():
            await self.inner_test_get_my_recent_trades()

    async def inner_test_get_my_recent_trades(self):
        trades = await self.get_my_recent_trades()
        assert trades
        self.check_raw_trades(trades)

    async def test_get_closed_orders(self):
        async with self.local_exchange_manager():
            await self.inner_test_get_closed_orders()

    async def inner_test_get_closed_orders(self):
        orders = await self.get_closed_orders()
        assert orders
        self.check_raw_closed_orders(orders)

    async def test_edit_limit_order(self):
        # pass if not implemented
        async with self.local_exchange_manager():
            await self.inner_test_edit_limit_order()

    async def inner_test_edit_limit_order(self):
        current_price = await self.get_price()
        portfolio = await self.get_portfolio()
        price = self.get_order_price(current_price, True)
        size = self.get_order_size(portfolio, price)
        open_orders = await self.get_open_orders()
        sell_limit = await self.create_limit_order(price, size, trading_enums.TradeOrderSide.SELL)
        self.check_created_limit_order(sell_limit, price, size, trading_enums.TradeOrderSide.SELL)
        assert await self.order_in_open_orders(open_orders, sell_limit)
        edited_price = self.get_order_price(current_price, True, price_diff=2*self.ORDER_PRICE_DIFF)
        edited_size = self.get_order_size(portfolio, price, order_size=2*self.ORDER_SIZE)
        await self.edit_order(sell_limit,
                              edited_price=edited_price,
                              edited_quantity=edited_size)
        await self.wait_for_edit(sell_limit, edited_size)
        sell_limit = await self.get_order(sell_limit.order_id)
        self.check_created_limit_order(sell_limit, edited_price, edited_size, trading_enums.TradeOrderSide.SELL)
        await self.cancel_order(sell_limit)
        assert await self.order_not_in_open_orders(open_orders, sell_limit)

    async def test_edit_stop_order(self):
        # pass if not implemented
        async with self.local_exchange_manager():
            await self.inner_test_edit_stop_order()

    async def inner_test_edit_stop_order(self):
        current_price = await self.get_price()
        portfolio = await self.get_portfolio()
        price = self.get_order_price(current_price, True)
        size = self.get_order_size(portfolio, price)
        open_orders = await self.get_open_orders()
        stop_loss = await self.create_market_stop_loss_order(current_price, price, size,
                                                             trading_enums.TradeOrderSide.SELL)
        self.check_created_stop_order(stop_loss, price, size, trading_enums.TradeOrderSide.SELL)
        assert await self.order_in_open_orders(open_orders, stop_loss)
        edited_price = self.get_order_price(current_price, True, price_diff=2*self.ORDER_PRICE_DIFF)
        edited_size = self.get_order_size(portfolio, price, order_size=2*self.ORDER_SIZE)
        await self.edit_order(stop_loss,
                              edited_stop_price=edited_price,
                              edited_quantity=edited_size)
        await self.wait_for_edit(stop_loss, edited_size)
        stop_loss = await self.get_order(stop_loss.order_id)
        self.check_created_stop_order(stop_loss, edited_price, edited_size, trading_enums.TradeOrderSide.SELL)
        await self.cancel_order(stop_loss)
        assert await self.order_not_in_open_orders(open_orders, stop_loss)

    async def test_create_bundled_orders(self):
        # pass if not implemented
        async with self.local_exchange_manager():
            await self.inner_test_create_bundled_orders()

    async def inner_test_create_bundled_orders(self):
        current_price = await self.get_price()
        stop_loss_price = self.get_order_price(current_price, False)
        take_profit_price = self.get_order_price(current_price, True)
        size = self.get_order_size(await self.get_portfolio(), stop_loss_price)
        open_orders = await self.get_open_orders()
        stop_loss = await self.create_market_stop_loss_order(current_price, stop_loss_price, size,
                                                             trading_enums.TradeOrderSide.SELL,
                                                             push_on_exchange=False)
        take_profit = await self.create_order(take_profit_price, current_price, size, trading_enums.TradeOrderSide.SELL,
                                              trading_enums.TraderOrderType.TAKE_PROFIT, push_on_exchange=False)
        market_order = await self.create_market_order(current_price, size,
                                                      trading_enums.TradeOrderSide.BUY,
                                                      push_on_exchange=False)
        # bundle stop loss and take profits into open position order
        params = await self.exchange_manager.trader.bundle_chained_order_with_uncreated_order(market_order, stop_loss)
        params.update(
            await self.exchange_manager.trader.bundle_chained_order_with_uncreated_order(market_order, take_profit)
        )
        buy_market = await self._create_order_on_exchange(market_order, params=params)
        self.check_created_market_order(buy_market, size, trading_enums.TradeOrderSide.BUY)
        await self.wait_for_fill(buy_market)
        created_orders = [stop_loss, take_profit]
        fetched_conditional_orders = await self.get_similar_orders_in_open_orders(open_orders, created_orders)
        for fetched_conditional_order in fetched_conditional_orders:
            # ensure stop loss / take profit is fetched in open orders
            # ensure stop loss / take profit cancel is working
            await self.cancel_order(fetched_conditional_order)
        for fetched_conditional_order in fetched_conditional_orders:
            assert await self.order_not_in_open_orders(open_orders, fetched_conditional_order)
        # close position
        sell_market = await self.create_market_order(current_price, size,
                                                     trading_enums.TradeOrderSide.SELL)
        self.check_created_market_order(sell_market, size, trading_enums.TradeOrderSide.SELL)
        await self.wait_for_fill(sell_market)

    async def get_portfolio(self):
        return await self.exchange_manager.exchange.get_balance()

    async def get_my_recent_trades(self, symbol=None):
        return await self.exchange_manager.exchange.get_my_recent_trades(symbol or self.SYMBOL)

    async def get_closed_orders(self, symbol=None):
        return await self.exchange_manager.exchange.get_closed_orders(symbol or self.SYMBOL)

    def check_duplicate(self, orders_or_trades):
        assert len({
            f"{o[trading_enums.ExchangeConstantsOrderColumns.ID.value]}"
            f"{o[trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value]}"
            f"{o[trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value]}"
            f"{o[trading_enums.ExchangeConstantsOrderColumns.PRICE.value]}"
            for o in orders_or_trades
        }) == len(orders_or_trades)

    def check_raw_closed_orders(self, closed_orders):
        self.check_duplicate(closed_orders)
        for closed_order in closed_orders:
            self.check_parsed_closed_order(
                personal_data.create_order_instance_from_raw(self.exchange_manager.trader, closed_order)
            )

    def check_parsed_closed_order(self, order: personal_data.Order):
        assert order.symbol
        assert order.timestamp
        assert order.order_type
        assert order.status
        if self.NO_FEE_ON_GET_CLOSED_ORDERS:
            assert order.fee is None
        else:
            assert order.fee
        assert order.order_id
        assert order.side
        if order.status not in (trading_enums.OrderStatus.REJECTED, trading_enums.OrderStatus.CANCELED):
            assert order.origin_quantity
            if self.OPEN_ORDERS_IN_CLOSED_ORDERS and order.status is trading_enums.OrderStatus.OPEN:
                # when order is open, cost is not full
                return
            self.check_theoretical_cost(
                symbols.parse_symbol(order.symbol), order.origin_quantity, order.origin_price, order.total_cost
            )

    def check_raw_trades(self, trades):
        self.check_duplicate(trades)
        for trade in trades:
            self.check_parsed_trade(
                personal_data.create_trade_instance_from_raw(self.exchange_manager.trader, trade)
            )

    def check_parsed_trade(self, trade: personal_data.Trade):
        assert trade.symbol
        assert trade.total_cost
        assert trade.trade_type
        assert trade.status
        assert trade.fee
        assert trade.origin_order_id
        assert trade.side
        if trade.status is not trading_enums.OrderStatus.CANCELED:
            assert trade.executed_quantity
            self.check_theoretical_cost(
                symbols.parse_symbol(trade.symbol), trade.executed_quantity,
                trade.executed_price, trade.total_cost
            )

    def check_theoretical_cost(self, symbol, quantity, price, cost):
        theoretical_cost = quantity * price
        assert theoretical_cost * decimal.Decimal("0.8") <= cost <= theoretical_cost * decimal.Decimal("1.2")

    async def get_price(self, symbol=None):
        return decimal.Decimal(str(
            (await self.exchange_manager.exchange.get_price_ticker(symbol or self.SYMBOL))[
                trading_enums.ExchangeConstantsTickersColumns.CLOSE.value
            ]
        ))

    @contextlib.asynccontextmanager
    async def local_exchange_manager(self):
        try:
            exchange_tentacle_name = self.EXCHANGE_TENTACLE_NAME or self.EXCHANGE_NAME.capitalize()
            async with get_authenticated_exchange_manager(
                    self.EXCHANGE_NAME,
                    exchange_tentacle_name,
                    self.get_config(),
                    credentials_exchange_name=self.CREDENTIALS_EXCHANGE_NAME or self.EXCHANGE_NAME
            ) as exchange_manager:
                self.exchange_manager = exchange_manager
                yield
        finally:
            self.exchange_manager = None

    def check_portfolio_content(self, portfolio):
        assert len(portfolio) >= self.MIN_PORTFOLIO_SIZE
        at_least_one_value = False
        for asset, values in portfolio.items():
            assert all(
                key in values
                for key in (
                    trading_constants.CONFIG_PORTFOLIO_FREE,
                    trading_constants.CONFIG_PORTFOLIO_USED,
                    trading_constants.CONFIG_PORTFOLIO_TOTAL
                )
            )
            if values[trading_constants.CONFIG_PORTFOLIO_TOTAL] > trading_constants.ZERO:
                at_least_one_value = True
        assert at_least_one_value

    async def get_order(self, order_id, symbol=None):
        return personal_data.create_order_instance_from_raw(
            self.exchange_manager.trader,
            await self.exchange_manager.exchange.get_order(order_id, symbol or self.SYMBOL)
        )

    async def edit_order(self, order,
                         edited_quantity: decimal.Decimal = None,
                         edited_price: decimal.Decimal = None,
                         edited_stop_price: decimal.Decimal = None,
                         edited_current_price: decimal.Decimal = None,
                         params: dict = None):
        edited_order = await self.exchange_manager.exchange.edit_order(
            order.order_id,
            order.order_type,
            order.symbol,
            quantity=order.origin_quantity if edited_quantity is None else edited_quantity,
            price=order.origin_price if edited_price is None else edited_price,
            stop_price=edited_stop_price,
            side=order.side,
            current_price=edited_current_price,
            params=params
        )
        assert edited_order is not None
        return edited_order

    async def create_limit_order(self, price, size, side, symbol=None,
                                 push_on_exchange=True):
        order_type = trading_enums.TraderOrderType.BUY_LIMIT \
            if side is trading_enums.TradeOrderSide.BUY else trading_enums.TraderOrderType.SELL_LIMIT
        return await self.create_order(price, price, size, side, order_type,
                                       symbol=symbol, push_on_exchange=push_on_exchange)

    async def create_market_stop_loss_order(self, current_price, stop_price, size, side, symbol=None,
                                            push_on_exchange=True):
        self.exchange_manager.trader.allow_artificial_orders = False
        return await self.create_order(stop_price, current_price, size, side, trading_enums.TraderOrderType.STOP_LOSS,
                                       symbol=symbol, push_on_exchange=push_on_exchange)

    async def create_market_order(self, current_price, size, side, symbol=None,
                                  push_on_exchange=True):
        order_type = trading_enums.TraderOrderType.BUY_MARKET \
            if side is trading_enums.TradeOrderSide.BUY else trading_enums.TraderOrderType.SELL_MARKET
        return await self.create_order(None, current_price, size, side, order_type,
                                       symbol=symbol, push_on_exchange=push_on_exchange)

    async def create_order(self, price, current_price, size, side, order_type,
                           symbol=None, push_on_exchange=True):
        current_order = personal_data.create_order_instance(
            self.exchange_manager.trader,
            order_type=order_type,
            symbol=symbol or self.SYMBOL,
            current_price=current_price,
            quantity=size,
            price=price,
            side=side,
        )
        if push_on_exchange:
            current_order = await self._create_order_on_exchange(current_order)
        if current_order is None:
            raise AssertionError("Error when creating order")
        return current_order

    async def _create_order_on_exchange(self, order, params=None):
        created_order = await self.exchange_manager.trader.create_order(order, params=params, wait_for_creation=False)
        if created_order.status is trading_enums.OrderStatus.PENDING_CREATION:
            await self.wait_for_open(created_order)
            return await self.get_order(created_order.order_id)
        return created_order

    def get_order_size(self, portfolio, price, symbol=None, order_size=None):
        order_size = order_size or self.ORDER_SIZE
        currency_quantity = portfolio[self.SETTLEMENT_CURRENCY][self.PORTFOLIO_TYPE_FOR_SIZE] \
            * decimal.Decimal(order_size) / trading_constants.ONE_HUNDRED
        symbol = symbols.parse_symbol(symbol or self.SYMBOL)
        if symbol.is_inverse():
            order_quantity = currency_quantity * price
        else:
            order_quantity = currency_quantity / price
        return personal_data.decimal_adapt_quantity(
            self.exchange_manager.exchange.get_market_status(str(symbol)),
            order_quantity
        )

    def get_sell_size_from_buy_order(self, buy_order):
        sell_size = buy_order.origin_quantity
        if buy_order.fee and buy_order.fee[trading_enums.FeePropertyColumns.CURRENCY.value] == self.ORDER_CURRENCY:
            sell_size = sell_size - buy_order.fee[trading_enums.FeePropertyColumns.COST.value]
        return personal_data.decimal_adapt_quantity(
            self.exchange_manager.exchange.get_market_status(str(buy_order.symbol)),
            sell_size
        )

    def get_order_price(self, price, is_above_price, symbol=None, price_diff=None):
        price_diff = price_diff or self.ORDER_PRICE_DIFF
        multiplier = 1 + price_diff / 100 if is_above_price else 1 - price_diff / 100
        return personal_data.decimal_adapt_price(
            self.exchange_manager.exchange.get_market_status(symbol or self.SYMBOL),
            price * (decimal.Decimal(str(multiplier)))
        )

    async def get_open_orders(self, symbol=None):
        orders = await self.exchange_manager.exchange.get_open_orders(symbol or self.SYMBOL)
        self.check_duplicate(orders)
        return orders

    def check_created_limit_order(self, order, price, size, side):
        self._check_order(order, size, side)
        assert order.origin_price == price
        expected_type = personal_data.BuyLimitOrder \
            if side is trading_enums.TradeOrderSide.BUY else personal_data.SellLimitOrder
        assert isinstance(order, expected_type)

    def check_created_market_order(self, order, size, side):
        self._check_order(order, size, side)
        expected_type = personal_data.BuyMarketOrder \
            if side is trading_enums.TradeOrderSide.BUY else personal_data.SellMarketOrder
        assert isinstance(order, expected_type)

    def check_created_stop_order(self, order, price, size, side):
        self._check_order(order, size, side)
        assert order.origin_price == price
        assert order.side is side
        expected_type = personal_data.StopLossOrder
        assert isinstance(order, expected_type)

    def check_portfolio_changed(self, previous_portfolio, updated_portfolio, has_increased, symbol=None):
        previous_free_quantity = previous_portfolio[symbol or self.SETTLEMENT_CURRENCY][
            trading_constants.CONFIG_PORTFOLIO_FREE]
        updated_free_quantity = updated_portfolio[symbol or self.SETTLEMENT_CURRENCY][
            trading_constants.CONFIG_PORTFOLIO_FREE]
        if has_increased:
            assert updated_free_quantity > previous_free_quantity
        else:
            assert updated_free_quantity < previous_free_quantity

    def _check_order(self, order, size, side):
        if self.CONVERTS_ORDER_SIZE_BEFORE_PUSHING_TO_EXCHANGES:
            # actual origin_quantity may vary due to quantity conversion for market orders
            assert size * decimal.Decimal("0.8") <= order.origin_quantity <= size * decimal.Decimal("1.2")
        else:
            assert order.origin_quantity == size
        assert order.side is side
        assert order.is_open()

    async def wait_for_fill(self, order):
        def parse_is_filled(raw_order):
            return personal_data.parse_order_status(raw_order) in {trading_enums.OrderStatus.FILLED,
                                                                   trading_enums.OrderStatus.CLOSED}
        await self._get_order_until(order, parse_is_filled, self.MARKET_FILL_TIMEOUT)

    def parse_order_is_not_pending(self, raw_order):
        return personal_data.parse_order_status(raw_order) not in (trading_enums.OrderStatus.UNKNOWN, None)

    async def wait_for_open(self, order):
        await self._get_order_until(order, self.parse_order_is_not_pending, self.OPEN_TIMEOUT)

    async def wait_for_cancel(self, order):
        return personal_data.create_order_instance_from_raw(
            self.exchange_manager.trader,
            await self._get_order_until(order, personal_data.parse_is_cancelled, self.CANCEL_TIMEOUT)
        )

    async def wait_for_edit(self, order, edited_quantity):
        def is_edited(row_order):
            return decimal.Decimal(str(row_order[trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value])) \
                   == edited_quantity
        await self._get_order_until(order, is_edited, self.EDIT_TIMEOUT)

    async def _get_order_until(self, order, validation_func, timeout):
        t0 = time.time()
        while time.time() - t0 < timeout:
            raw_order = await self.exchange_manager.exchange.get_order(order.order_id, order.symbol)
            if raw_order and validation_func(raw_order):
                return raw_order
        raise TimeoutError(f"Order not filled within {timeout}s: {order}")

    async def order_in_open_orders(self, previous_open_orders, order):
        open_orders = await self.get_open_orders()
        assert len(open_orders) == len(previous_open_orders) + 1
        for open_order in open_orders:
            if open_order[trading_enums.ExchangeConstantsOrderColumns.ID.value] == order.order_id:
                return True
        return False

    async def get_similar_orders_in_open_orders(self, previous_open_orders, orders):
        if not isinstance(orders, list):
            orders = [orders]
        found_orders = []
        open_orders = await self.get_open_orders()
        assert len(open_orders) == len(previous_open_orders) + len(orders)
        for order in orders:
            for open_order in open_orders:
                fetched_order = personal_data_orders.order_factory.create_order_instance_from_raw(
                    self.exchange_manager.trader,
                    open_order
                )
                if personal_data.is_associated_pending_order(order, fetched_order):
                    found_orders.append(fetched_order)
                    break
        if len(found_orders) == len(orders):
            return found_orders
        raise AssertionError(f"Can't find any order similar to {orders}. Found: {found_orders}")

    async def order_not_in_open_orders(self, previous_open_orders, order):
        open_orders = await self.get_open_orders()
        assert len(open_orders) == len(previous_open_orders)
        for open_order in open_orders:
            if open_order[trading_enums.ExchangeConstantsOrderColumns.ID.value] == order.order_id:
                return False
        return True

    async def cancel_order(self, order):
        cancelled_order = order
        if not await self.exchange_manager.trader.cancel_order(order, wait_for_cancelling=False):
            raise AssertionError("cancel_order returned False")
        if order.status is trading_enums.OrderStatus.PENDING_CANCEL:
            cancelled_order = await self.wait_for_cancel(order)
        assert cancelled_order.status is trading_enums.OrderStatus.CANCELED
        if cancelled_order.state is not None:
            assert cancelled_order.is_cancelled()
        return order

    def get_config(self):
        return {
            constants.CONFIG_EXCHANGES: {
                self.EXCHANGE_NAME: {
                    constants.CONFIG_EXCHANGE_TYPE: self.EXCHANGE_TYPE
                }
            },
            constants.CONFIG_CRYPTO_CURRENCIES: {
                "test-crypto": {
                    constants.CONFIG_CRYPTO_PAIRS: self._get_all_symbols()
                }
            }
        }

    def _get_all_symbols(self):
        return [self.SYMBOL]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import contextlib
import os
import dotenv
import mock

import trading_backend
import octobot_commons.constants as commons_constants
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.os_util as os_util
import octobot_commons.tests.test_config as test_config
import octobot_trading.api as trading_api
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as trading_constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.personal_data as personal_data
import octobot_tentacles_manager.constants as tentacles_manager_constants
import tests.test_utils.config as test_utils_config


LOADED_EXCHANGE_CREDS_ENV_VARIABLES = False


class ExchangeChannelMock:
    def __init__(self, exchange_manager, name):
        self.exchange_manager = exchange_manager
        self.name = name

        self.get_internal_producer = mock.Mock(
            return_value=mock.Mock(
                update_order_from_exchange=mock.AsyncMock(),
                send=mock.AsyncMock(),
            )
        )
        self.get_consumers = mock.Mock(return_value=[mock.Mock()])

    def get_name(self):
        return self.name


@contextlib.asynccontextmanager
async def get_authenticated_exchange_manager(exchange_name, exchange_tentacle_name, config=None,
                                             credentials_exchange_name=None):
    credentials_exchange_name = credentials_exchange_name or exchange_name
    _load_exchange_creds_env_variables_if_necessary()
    config = {**test_config.load_test_config(), **config} if config else test_config.load_test_config()
    if exchange_name not in config[commons_constants.CONFIG_EXCHANGES]:
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {}
    config[commons_constants.CONFIG_EXCHANGES][exchange_name].update(_get_exchange_auth_details(
        credentials_exchange_name
    ))
    exchange_type = config[commons_constants.CONFIG_EXCHANGES][exchange_name].get(
        commons_constants.CONFIG_EXCHANGE_TYPE, exchanges.get_default_exchange_type(exchange_name))
    exchange_builder = trading_api.create_exchange_builder(config, exchange_name) \
        .has_matrix("") \
        .use_tentacles_setup_config(get_tentacles_setup_config_with_exchange(exchange_tentacle_name)) \
        .set_bot_id("") \
        .is_real() \
        .is_checking_credentials(False) \
        .is_sandboxed(_get_exchange_is_sandboxed(exchange_name)) \
        .is_using_exchange_type(exchange_type) \
        .enable_storage(False) \
        .disable_trading_mode() \
        .is_exchange_only()
    exchange_manager_instance = await exchange_builder.build()
    # create trader afterwards to init exchange personal data
    exchange_manager_instance.trader.is_enabled = True
    await exchange_manager_instance.register_trader(exchange_manager_instance.trader)
    exchange_manager_instance.exchange_backend = trading_backend.exchange_factory.create_exchange_backend(
        exchange_manager_instance.exchange
    )
    exchange_manager_instance.exchange.__class__.PRINT_DEBUG_LOGS = True
    set_mocked_required_channels(exchange_manager_instance)
    try:
        yield exchange_manager_instance
    except errors.UnreachableExchange as err:
        raise errors.UnreachableExchange(f"{exchange_name} can't be reached, it is either offline or you are not connected "
                                         "to the internet (or a proxy is preventing connecting to this exchange).") \
            from err
    finally:
        await exchange_manager_instance.stop()
        trading_api.cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await asyncio_tools.wait_asyncio_next_cycle()


def set_mocked_required_channels(exchange_manager):
    # disable waiting time as order refresh is mocked
    personal_data.State.PENDING_REFRESH_INTERVAL = 0
    for channel in (trading_constants.ORDERS_CHANNEL, trading_constants.BALANCE_CHANNEL):
        exchange_channel.set_chan(ExchangeChannelMock(exchange_manager, channel), channel)


def get_tentacles_setup_config_with_exchange(exchange_tentacle_name):
    setup_config = test_utils_config.load_test_tentacles_config()
    setup_config.tentacles_activation[tentacles_manager_constants.TENTACLES_TRADING_PATH][exchange_tentacle_name] = True
    return setup_config


def _load_exchange_creds_env_variables_if_necessary():
    global LOADED_EXCHANGE_CREDS_ENV_VARIABLES
    if not LOADED_EXCHANGE_CREDS_ENV_VARIABLES:
        # load environment variables from .env file if exists
        dotenv_path = os.getenv("EXCHANGE_TESTS_DOTENV_PATH", os.path.dirname(os.path.abspath(__file__)))
        dotenv.load_dotenv(os.path.join(dotenv_path, ".env"), verbose=False)
        LOADED_EXCHANGE_CREDS_ENV_VARIABLES = True


def _get_exchange_auth_details(exchange_name):
    return {
        commons_constants.CONFIG_EXCHANGE_KEY:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_KEY),
        commons_constants.CONFIG_EXCHANGE_SECRET:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_SECRET),
        commons_constants.CONFIG_EXCHANGE_PASSWORD:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_PASSWORD),
    }


def _get_exchange_credential_from_env(exchange_name, cred_suffix):
    # for bybit api key: get BYBIT_KEY (as encrypted value)
    # for bybit api password: get BYBIT_PASSWORD (as encrypted value)
    # for bybit api secret: get BYBIT_SECRET (as encrypted value)
    return os.getenv(f"{exchange_name}_{cred_suffix.split('-')[-1]}".upper(), None)


def _get_exchange_is_sandboxed(exchange_name):
    # for bybit api secret: get BYBIT_SANDBOXED (as true / false)
    return os_util.parse_boolean_environment_var(f"{exchange_name}_SANDBOXED".upper(), "false")



# cython: language_level=3
#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot.configuration_manager as configuration_manager
cimport octobot.octobot_channel_consumer as octobot_channel_consumer
cimport octobot.initializer as initializer
cimport octobot.octobot_api as octobot_api
cimport octobot.producers as producers
cimport octobot.task_manager as task_manager


cdef class OctoBot:
    cdef object logger

    cdef public double start_time

    cdef public bint reset_trading_history
    cdef public bint initialized
    cdef public bint ignore_config
    cdef public list startup_messages

    cdef public dict tools
    cdef public dict config
    cdef public str bot_id

    cdef object _aiohttp_session
    cdef public object community_handler
    cdef public object community_auth
    cdef public object async_loop
    cdef public object tentacles_setup_config
    cdef public object stopped  # asyncio.Event
    cdef public object automation

    cdef public initializer.Initializer initializer
    cdef public task_manager.TaskManager task_manager
    cdef object _init_metadata_run_task
    cdef public producers.ExchangeProducer exchange_producer
    cdef public producers.EvaluatorProducer evaluator_producer
    cdef public producers.InterfaceProducer interface_producer
    cdef public producers.ServiceFeedProducer service_feed_producer
    cdef public octobot_api.OctoBotAPI octobot_api
    cdef public octobot_channel_consumer.OctoBotChannelGlobalConsumer global_consumer
    cdef public configuration_manager.ConfigurationManager configuration_manager

    cpdef object run_in_main_asyncio_loop(self, object coroutine, bint log_exceptions=*, object timeout=*)
    cpdef void set_watcher(self, object watcher)
    cpdef object get_aiohttp_session(self)
    cpdef object get_edited_config(self, str config_key, bint dict_only=*)
    cpdef void set_edited_config(self, str config_key, object config)
    cpdef object get_startup_config(self, str config_key, bint dict_only=*)
    cpdef object get_trading_mode(self)
